# src/dynamic_sarimax/metrics.py
from __future__ import annotations
import numpy as np


def mse(y_true, y_pred) -> float:
    y = np.asarray(y_true, dtype=float)
    yhat = np.asarray(y_pred, dtype=float)
    return float(np.mean((y - yhat) ** 2))


def smape(y_true, y_pred) -> float:
    y = np.asarray(y_true, dtype=float)
    yhat = np.asarray(y_pred, dtype=float)
    denom = np.abs(y) + np.abs(yhat)
    denom = np.where(denom == 0, 1.0, denom)
    return float(np.mean(200.0 * np.abs(y - yhat) / denom))



# src/dynamic_sarimax/features.py
from __future__ import annotations
from dataclasses import dataclass
import numpy as np
import pandas as pd
from .config import ExogLagSpec
from .utils import ensure_dataframe, check_monotone_index, safe_shift, finite_mask_df


@dataclass
class _Scaler:
    mean_: float | None = None
    std_: float | None = None

    def fit(self, s: pd.Series) -> None:
        self.mean_ = float(np.nanmean(s.values))
        sd = float(np.nanstd(s.values, ddof=1))
        self.std_ = 1.0 if (not np.isfinite(sd) or sd == 0.0) else sd

    def transform(self, s: pd.Series) -> pd.Series:
        if self.mean_ is None or self.std_ is None:
            raise RuntimeError("Scaler not fit")
        return (s - self.mean_) / self.std_


class ExogLagTransformer:
    """Safe lag + train-only scaling with strict alignment."""

    def __init__(self, spec: ExogLagSpec):
        if spec.delay < 0:
            raise ValueError("delay must be >= 0")
        self.spec = spec
        self._scalers: dict[str, _Scaler] = {}
        self._mask: pd.Series | None = None
        self._columns: list[str] | None = None
        self._fitted_index: pd.Index | None = None

    def _lag_all(self, Xdf: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            {col: safe_shift(Xdf[col], self.spec.delay) for col in Xdf.columns},
            index=Xdf.index,
        )

    def fit(self, X: pd.Series | pd.DataFrame, y: pd.Series) -> "ExogLagTransformer":
        Xdf = ensure_dataframe(X, name="X")
        if Xdf is None:
            raise ValueError("Exogenous X required for ExogLagTransformer.fit")
        check_monotone_index(Xdf, "X")
        check_monotone_index(y, "y")
        if not Xdf.index.equals(y.index):
            common = Xdf.index.intersection(y.index)
            Xdf = Xdf.loc[common]
            y = y.loc[common]
        Xlag = self._lag_all(Xdf)
        mask = finite_mask_df(Xlag)
        self._mask = mask
        self._columns = list(Xdf.columns)
        self._fitted_index = Xdf.index

        if self.spec.scale:
            for col in Xlag.columns:
                sc = _Scaler()
                sc.fit(Xlag.loc[mask, col])
                self._scalers[col] = sc
        return self

    def transform(self, X: pd.Series | pd.DataFrame) -> pd.DataFrame:
        if self._columns is None or self._mask is None:
            raise RuntimeError("Transformer not fit")
        Xdf = ensure_dataframe(X, name="X")
        if list(Xdf.columns) != self._columns:
            Xdf = Xdf.copy()
            Xdf.columns = self._columns
        Xlag = self._lag_all(Xdf)
        if self.spec.scale:
            for col, sc in self._scalers.items():
                Xlag[col] = sc.transform(Xlag[col])
        return Xlag

    def mask(self) -> pd.Series:
        if self._mask is None:
            raise RuntimeError("Transformer not fit")
        return self._mask.copy()

    def trim_target(self, y: pd.Series) -> pd.Series:
        if self._mask is None:
            raise RuntimeError("Transformer not fit")
        return y.loc[self._mask]

    def future_block(
        self, X_full: pd.DataFrame, start_idx: int, steps: int
    ) -> pd.DataFrame:
        """Return lagged+scaled exog for indices [start_idx, start_idx+steps-1]."""
        if self._columns is None:
            raise RuntimeError("Transformer not fit")
        Xdf = ensure_dataframe(X_full, name="X_full")
        if Xdf.shape[1] != len(self._columns):
            raise ValueError(
                f"Future exogenous column count mismatch. Expected {len(self._columns)}, got {Xdf.shape[1]}"
            )
        if list(Xdf.columns) != self._columns:
            Xdf = Xdf.copy()
            Xdf.columns = self._columns
        Xlag = self._lag_all(Xdf)
        if self.spec.scale:
            for col, sc in self._scalers.items():
                Xlag[col] = sc.transform(Xlag[col])
        block = Xlag.iloc[start_idx : start_idx + steps]
        if not np.all(np.isfinite(block.values)):
            raise ValueError(
                "future_block contains NaN/Inf - not enough history for lagged exog at these horizons"
            )
        return block



# src/dynamic_sarimax/model.py
from __future__ import annotations
from typing import Optional, Dict, Any
import pandas as pd
from statsmodels.tsa.statespace.sarimax import SARIMAX
from .config import SarimaxConfig
from .features import ExogLagTransformer
from .utils import check_monotone_index


class DynamicSarimax:
    """SARIMAX wrapper that refuses unsafe exogenous alignment."""

    def __init__(self, cfg: SarimaxConfig, lagger: Optional[ExogLagTransformer] = None):
        self.cfg = cfg
        self.lagger = lagger
        self._fit_res = None

    def fit(self, y: pd.Series, X: Optional[pd.DataFrame] = None) -> "DynamicSarimax":
        check_monotone_index(y, "y")
        exog = None
        y_use = y
        if self.lagger is not None:
            if X is None:
                raise ValueError("Exogenous X required when lagger is set")
            check_monotone_index(X, "X")
            if not X.index.equals(y.index):
                common = X.index.intersection(y.index)
                X = X.loc[common]
                y_use = y.loc[common]
            self.lagger.fit(X, y_use)
            X_lag = self.lagger.transform(X)
            mask = self.lagger.mask()
            exog = X_lag.loc[mask]
            y_use = y_use.loc[mask]
        trend = self.cfg.materialize_trend()
        self._fit_res = SARIMAX(
            y_use.values,
            order=self.cfg.order,
            seasonal_order=self.cfg.seasonal_order,
            exog=None if exog is None else exog.values,
            trend=trend,
            enforce_stationarity=self.cfg.enforce_stationarity,
            enforce_invertibility=self.cfg.enforce_invertibility,
        ).fit(disp=False)
        return self

    def forecast(
        self,
        steps: int,
        X_future: Optional[pd.DataFrame] = None,
        start_idx: Optional[int] = None,
    ) -> pd.Series:
        if self._fit_res is None:
            raise RuntimeError("Model not fit")
        if self.lagger is None:
            pred = self._fit_res.get_forecast(steps=steps)
            return pd.Series(pred.predicted_mean)
        # with exog: require future exog aligned via lagger
        if X_future is None or start_idx is None:
            raise ValueError(
                "X_future and start_idx required when using exogenous lagger"
            )
        block = self.lagger.future_block(X_future, start_idx=start_idx, steps=steps)
        pred = self._fit_res.get_forecast(steps=steps, exog=block.values)
        return pd.Series(pred.predicted_mean)

    def aic(self) -> float:
        if self._fit_res is None:
            raise RuntimeError("Model not fit")
        return float(self._fit_res.aic)

    def model_info(self) -> Dict[str, Any]:
        if self._fit_res is None:
            raise RuntimeError("Model not fit")
        return {
            "order": self.cfg.order,
            "seasonal_order": self.cfg.seasonal_order,
            "aic": float(self._fit_res.aic),
            "nobs": int(self._fit_res.nobs),
        }



# src/dynamic_sarimax/utils.py
from __future__ import annotations
import numpy as np
import pandas as pd


def ensure_series(
    x: pd.Series | pd.DataFrame | np.ndarray, name: str = "x"
) -> pd.Series:
    if isinstance(x, pd.Series):
        return x
    if isinstance(x, pd.DataFrame):
        if x.shape[1] != 1:
            raise ValueError(f"{name} must be 1-D when passed here; got {x.shape}")
        return x.iloc[:, 0]
    if isinstance(x, np.ndarray):
        if x.ndim != 1:
            raise ValueError(f"{name} must be 1-D array; got shape {x.shape}")
        return pd.Series(x)
    raise TypeError(f"Unsupported type for {name}: {type(x)}")


def ensure_dataframe(
    X: pd.Series | pd.DataFrame | None, name: str = "X"
) -> pd.DataFrame | None:
    if X is None:
        return None
    if isinstance(X, pd.DataFrame):
        return X
    if isinstance(X, pd.Series):
        col = X.name if (X.name is not None and X.name != 0) else "exog"
        return X.to_frame(name=col)
    raise TypeError(f"Unsupported type for {name}: {type(X)}")


def check_monotone_index(obj: pd.Series | pd.DataFrame, name: str) -> None:
    idx = obj.index
    if not idx.is_monotonic_increasing:
        raise ValueError(f"{name} index must be monotone increasing")
    if idx.has_duplicates:
        raise ValueError(f"{name} index has duplicates")


def safe_shift(s: pd.Series, b: int) -> pd.Series:
    if b < 0:
        raise ValueError("delay must be >= 0")
    return s.shift(b)


def finite_mask_df(df: pd.DataFrame) -> pd.Series:
    m = np.all(np.isfinite(df.values), axis=1)
    return pd.Series(m, index=df.index)



# src/dynamic_sarimax/evaluation.py
from __future__ import annotations
import numpy as np
import pandas as pd
from typing import Optional
from .config import SarimaxConfig, ExogLagSpec
from .features import ExogLagTransformer
from .model import DynamicSarimax
from .metrics import mse, smape


def rolling_evaluate(
    y: pd.Series,
    X: Optional[pd.Series | pd.DataFrame],
    cfg: SarimaxConfig,
    delay: Optional[int],
    horizons: int,
    train_frac: float = 0.8,
    min_train: int = 30,
) -> pd.DataFrame:
    N = len(y)
    train_end0 = int(np.floor(train_frac * N))
    test_start = train_end0
    last_origin = N - horizons
    if last_origin < test_start:
        raise ValueError("horizons too large relative to series length")

    rows = []
    Xdf = (
        None
        if X is None
        else (
            X
            if isinstance(X, pd.DataFrame)
            else X.to_frame(name=getattr(X, "name", "exog"))
        )
    )

    for o in range(test_start, last_origin + 1):
        y_tr = y.iloc[:o]
        X_tr = None if Xdf is None else Xdf.iloc[:o]
        if delay is None:
            model = DynamicSarimax(cfg=cfg, lagger=None)
            model.fit(y_tr, None)
            steps = horizons
            yhat = model.forecast(steps=steps)
            ytrue = y.iloc[o : o + steps].reset_index(drop=True)
            for h in range(steps):
                rows.append(
                    {
                        "origin": o,
                        "h": h + 1,
                        "y_true": float(ytrue.iloc[h]),
                        "y_hat": float(yhat.iloc[h]),
                    }
                )
            continue

        lagger = ExogLagTransformer(ExogLagSpec(delay=delay, scale=True))
        model = DynamicSarimax(cfg=cfg, lagger=lagger)
        model.fit(y_tr, X_tr)
        if model._fit_res.nobs < min_train:
            continue

        X_full = Xdf
        yhat = model.forecast(steps=horizons, X_future=X_full, start_idx=o)
        ytrue = y.iloc[o : o + horizons].reset_index(drop=True)
        for h in range(horizons):
            rows.append(
                {
                    "origin": o,
                    "h": h + 1,
                    "y_true": float(ytrue.iloc[h]),
                    "y_hat": float(yhat.iloc[h]),
                }
            )

    out = pd.DataFrame(rows)
    if out.empty:
        raise RuntimeError("No evaluations produced - check inputs")

    agg = (
        out.groupby("h", as_index=False)
        .apply(
            lambda g: pd.Series(
                {
                    "n_origins": len(g),
                    "MSE": mse(g["y_true"].values, g["y_hat"].values),
                    "sMAPE": smape(g["y_true"].values, g["y_hat"].values),
                }
            ),
            include_groups=False,
        )
        .reset_index(drop=True)
    )

    macro_MSE = agg["MSE"].mean()
    macro_sMAPE = agg["sMAPE"].mean()

    agg.attrs["macro_MSE"] = float(macro_MSE)
    agg.attrs["macro_sMAPE"] = float(macro_sMAPE)
    return agg



# src/dynamic_sarimax/__init__.py
from .config import SarimaxConfig, ExogLagSpec
from .features import ExogLagTransformer
from .model import DynamicSarimax
from .selection import select_delay_by_aic
from .evaluation import rolling_evaluate
from .metrics import mse, smape


__all__ = [
    "SarimaxConfig",
    "ExogLagSpec",
    "ExogLagTransformer",
    "DynamicSarimax",
    "select_delay_by_aic",
    "rolling_evaluate",
    "mse",
    "smape",
]



# src/dynamic_sarimax/selection.py
from __future__ import annotations
from typing import Sequence, Tuple
import pandas as pd
from .config import SarimaxConfig, ExogLagSpec
from .features import ExogLagTransformer
from .model import DynamicSarimax


def select_delay_by_aic(
    y_train: pd.Series,
    X_train: pd.Series | pd.DataFrame,
    delays: Sequence[int],
    cfg: SarimaxConfig,
    min_train: int = 30,
) -> Tuple[int, float]:
    best = (None, float("inf"))
    for b in delays:
        lagger = ExogLagTransformer(ExogLagSpec(delay=b, scale=True))
        model = DynamicSarimax(cfg=cfg, lagger=lagger)
        # fit will trim via lagger; if effective sample too small, skip
        model.fit(y_train, X_train)
        nobs = model._fit_res.nobs  # safe here
        if nobs < min_train:
            continue
        aic = model.aic()
        if aic < best[1]:
            best = (b, aic)
    if best[0] is None:
        raise RuntimeError(
            "AIC selection failed - not enough effective training data for given delays"
        )
    return best  # (best_delay, best_aic)



# src/dynamic_sarimax/config.py
from __future__ import annotations
from dataclasses import dataclass
from typing import Tuple


@dataclass(frozen=True)
class SarimaxConfig:
    order: Tuple[int, int, int]
    seasonal_order: Tuple[int, int, int, int]
    trend: str | None = "auto"  # "auto" -> "c" if d=D=0 else "n"
    enforce_stationarity: bool = False
    enforce_invertibility: bool = False

    def materialize_trend(self) -> str:
        d = self.order[1]
        D = self.seasonal_order[1]
        if self.trend == "auto":
            return "c" if (d == 0 and D == 0) else "n"
        return "n" if self.trend is None else self.trend


@dataclass(frozen=True)
class ExogLagSpec:
    delay: int  # b >= 0
    scale: bool = True
